!
! Module to interface the CUDA runtime functions
!

module cuda_runtime

        integer,parameter:: cudaHostAllocPortable=1, &
                            cudaHostAllocMapped= 2, &
                            cudaDeviceMapHost=8

  interface
!
! cudaHostAlloc
!
	integer (C_INT) function cudaHostAlloc(buffer, size ,flag)  bind(C,name="cudaHostAlloc") 
	 use ISO_C_BINDING
	 implicit none 
	 type (C_PTR)  :: buffer 
	 integer (C_SIZE_T), value :: size
	 integer (C_INT), value :: flag
       end function cudaHostAlloc
!
! cudaFreeHost
!
	 integer (C_INT) function cudaFreeHost(buffer)  bind(C,name="cudaFreeHost") 
	  use iso_c_binding 
	  implicit none 
	  type (C_PTR), value :: buffer
	 end function cudaFreeHost
!
! cudaSetDeviceFlag
!
	integer (C_INT) function cudaSetDeviceFlags(flag) bind(C,name="cudaSetDeviceFlags")
	 use iso_c_binding
	 implicit none 
	 integer (C_INT), value :: flag
	end function cudaSetDeviceFlags
 
  end interface
end module cuda_runtime
!
! main.f90
!
program main

       use iso_c_binding
       use cuda_runtime
       implicit none

       integer, parameter :: fp_kind = kind(0.0d0) ! Double precision
 
       real(fp_kind) ,pointer, dimension (:) :: A,C
       real(fp_kind) ,allocatable, dimension (:) :: B
       type(C_PTR)::cptr_A,cptr_C

       integer i, N, seed
       integer err

!  Number of elements in the  arrays
       N=10

! Initialize the random number generator
       seed=1
       call random_seed(seed)

! Allocate A and C using cudaHostAlloc and then map the C pointer to Fortran arrays

       write(*,*)'Allocate host memory'
       err=cudaSetDeviceFlags(cudaDeviceMapHost)
       if (err > 0) print *,"Error in setting cudaSetDeviceFlags=",err

       err = cudaHostAlloc(cptr_A,N*sizeof(fp_kind),cudaHostAllocMapped)
       if (err > 0) print *,"Error in allocating A with cuda HostAlloc =",err
       call c_f_pointer(cptr_A,A,(/N/))

       err = cudaHostAlloc(cptr_C,N*sizeof(fp_kind),cudaHostAllocMapped)
       if (err > 0) print *,"Error in allocating C with cuda HostAlloc =",err
       call c_f_pointer(cptr_C,C,(/N/))

! From this point on, we can use A and C as normal Fortran array


! Allocate B using standard allocate call
       allocate(B(N))

! Initialize A with random numbers
        call random_number(A)


! computing  the reference solution on the CPU
       write(*,*)'computation on CPU'
               do i = 1, N
                       B(i) = dexp(A(i))
               enddo

!  same computation on the GPU
       write(*,*)'computation on GPU'
               call gexp(A,C,N)

! Print the computed quantities
        do i = 1, N
          write (*,'(i2,1x,4(g14.8))'),i,A(i),B(i),C(i),C(i)-B(i)
         enddo

! Release memory
       deallocate(B)
       err = cudaFreeHost (cptr_A)
       err = cudaFreeHost (cptr_C)

end program Main

